"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
Object.defineProperty(exports, "AnalysisError", {
  enumerable: true,
  get: function () {
    return _error.AnalysisError;
  }
});
exports.analyzeBody = analyzeBody;
Object.defineProperty(exports, "analyzeHead", {
  enumerable: true,
  get: function () {
    return _head.analyzeHead;
  }
});
exports.needsProps = needsProps;
var _error = require("./analysis/error.js");
var _this_fields = require("./analysis/this_fields.js");
var _state = require("./analysis/state.js");
var _utils = require("./utils.js");
var _prop = require("./analysis/prop.js");
var _head = require("./analysis/head.js");
const SPECIAL_MEMBER_NAMES = new Set([
// Special variables
"context", "props", "refs", "state",
// Lifecycle
"constructor", "render", "componentDidCatch", "componentDidMount", "componentDidUpdate", "componentWillMount", "UNSAFE_componentWillMount", "componentWillReceiveProps", "UNSAFE_componentWillReceiveProps", "componentWillUpdate", "UNSAFE_componentWillUpdate", "componentWillUnmount",
// Lifecycle predicates
"shouldComponentUpdate", "getSnapshotBeforeUpdate", "getChildContext",
// APIs (including deprecated)
"isReactComponent", "isMounted", "forceUpdate", "setState", "replaceState"]);
const SPECIAL_STATIC_NAMES = new Set(["childContextTypes", "contextTypes", "contextType", "defaultProps", "getDerivedStateFromError", "getDerivedStateFromProps"]);
function analyzeBody(path, babel) {
  const {
    thisFields: sites,
    staticFields
  } = (0, _this_fields.analyzeThisFields)(path);
  const propsObjSites = sites.get("props") ?? [];
  sites.delete("props");
  const defaultPropsObjSites = staticFields.get("defaultProps") ?? [];
  staticFields.delete("defaultProps");
  const stateObjSites = sites.get("state") ?? [];
  sites.delete("state");
  const setStateSites = sites.get("setState") ?? [];
  sites.delete("setState");
  const states = (0, _state.analyzeState)(stateObjSites, setStateSites);
  const locals = analyzeOuterCapturings(path);
  let renderPath = undefined;
  const members = new Map();
  const thisRefs = [];
  for (const [name, fieldSites] of sites.entries()) {
    if (name === "render") {
      if (fieldSites.some(site => site.type === "expr")) {
        throw new _error.AnalysisError(`do not use this.render`);
      }
      const init = fieldSites.find(site => site.init);
      if (init) {
        if (init.path.isClassMethod()) {
          renderPath = init.path;
        }
      }
    } else if (!SPECIAL_MEMBER_NAMES.has(name)) {
      const init = fieldSites.find(site => site.init);
      if (init) {
        const init_ = init.init;
        if ((0, _utils.isClassMethodLike)(init.path)) {
          members.set(name, analyzeMethod(init.path));
        } else if (init_.type === "init_value") {
          const initPath = init_.valuePath;
          if (initPath.isFunctionExpression() || initPath.isArrowFunctionExpression()) {
            members.set(name, analyzeFuncDef(initPath));
          } else {
            throw new _error.AnalysisError(`Non-analyzable initialization of ${name}`);
          }
        } else {
          throw new _error.AnalysisError(`Non-analyzable initialization of ${name}`);
        }
      }
      for (const site of fieldSites) {
        if (site.init) {
          continue;
        }
        if (site.type !== "expr" || site.hasWrite) {
          throw new _error.AnalysisError(`Invalid use of this.${name}`);
        }
        thisRefs.push({
          kind: "userDefined",
          path: site.path,
          name
        });
      }
    } else {
      throw new _error.AnalysisError(`Cannot transform ${name}`);
    }
  }
  for (const [name] of staticFields) {
    if (!SPECIAL_STATIC_NAMES.has(name)) {
      throw new _error.AnalysisError(`Cannot transform static ${name}`);
    } else {
      throw new _error.AnalysisError(`Cannot transform static ${name}`);
    }
  }
  if (!renderPath) {
    throw new _error.AnalysisError(`Missing render method`);
  }
  const props = (0, _prop.analyzeProps)(propsObjSites, defaultPropsObjSites);
  for (const [name, propAnalysis] of props.props) {
    if (propAnalysis.needsAlias) {
      propAnalysis.newAliasName = newLocal(name, babel, locals);
    }
  }
  const render = analyzeRender(renderPath, babel, locals, props);
  const stateMap = new Map();
  for (const [name, stateAnalysis] of states.entries()) {
    const field = ensureState(stateMap, name, babel, locals);
    if (stateAnalysis.init) {
      field.init = stateAnalysis.init.valuePath;
    }
    for (const site of stateAnalysis.sites) {
      if (site.type === "state_init") {
        // do nothing
      } else if (site.type === "expr") {
        thisRefs.push({
          kind: "state",
          path: site.path,
          field
        });
      } else {
        thisRefs.push({
          kind: "setState",
          path: site.path,
          field,
          rhs: site.valuePath
        });
      }
    }
  }
  return {
    render,
    state: stateMap,
    members,
    thisRefs,
    props
  };
}
function analyzeRender(path, babel, locals, props) {
  const renames = [];
  for (const [name, binding] of Object.entries(path.scope.bindings)) {
    if (props.allAliases.some(alias => alias.scope === binding.scope && alias.localName === name)) {
      // Already handled as a prop alias
      continue;
    }
    const newName = newLocal(name, babel, locals);
    renames.push({
      scope: binding.scope,
      oldName: name,
      newName
    });
  }
  return {
    path,
    renames
  };
}
function analyzeMethod(path) {
  return {
    type: "method",
    path
  };
}
function analyzeFuncDef(initPath) {
  return {
    type: "func_def",
    initPath
  };
}
function analyzeOuterCapturings(classPath) {
  const capturings = new Set();
  function visitIdent(path) {
    path.getOuterBindingIdentifiers;
    const binding = path.scope.getBinding(path.node.name);
    if (!binding || binding.path.isAncestor(classPath)) {
      capturings.add(path.node.name);
    }
  }
  classPath.get("body").traverse({
    Identifier(path) {
      if (path.isReferencedIdentifier()) {
        visitIdent(path);
      }
    },
    JSXIdentifier(path) {
      if (path.isReferencedIdentifier()) {
        visitIdent(path);
      }
    }
  });
  return capturings;
}
function ensureState(state, name, babel, locals) {
  if (!state.has(name)) {
    const localName = newLocal(name, babel, locals);
    const localSetterName = newLocal(`set${name.replace(/^[a-z]/, s => s.toUpperCase())}`, babel, locals);
    state.set(name, {
      localName,
      localSetterName
    });
  }
  return state.get(name);
}
function newLocal(baseName, babel, locals) {
  let name = baseName.replace(/[^\p{ID_Continue}$\u200C\u200D]/gu, "");
  if (!/^[\p{ID_Start}_$]/u.test(name) || !babel.types.isValidIdentifier(name)) {
    name = `_${name}`;
  }
  if (locals.has(name)) {
    name = name.replace(/\d+$/, "");
    for (let i = 0;; i++) {
      if (i >= 1000000) {
        throw new Error("Unexpected infinite loop");
      }
      if (!locals.has(`${name}${i}`)) {
        name = `${name}${i}`;
        break;
      }
    }
  }
  locals.add(name);
  return name;
}
function needsProps(body) {
  return body.props.sites.length > 0;
}