"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.analyzeState = analyzeState;
var _utils = require("../utils.js");
var _error = require("./error.js");
function analyzeState(stateObjSites, setStateSites) {
  const states = new Map();
  function getState(name) {
    if (!states.has(name)) {
      states.set(name, {
        sites: []
      });
    }
    return states.get(name);
  }
  const init = stateObjSites.find(site => site.init);
  if (init) {
    const init_ = init.init;
    if (init_.type !== "init_value") {
      throw new _error.AnalysisError("Non-analyzable state initializer");
    }
    const initPath = init_.valuePath;
    if (!initPath.isObjectExpression()) {
      throw new _error.AnalysisError("Non-analyzable state initializer");
    }
    for (const fieldPath of initPath.get("properties")) {
      if (!fieldPath.isObjectProperty()) {
        throw new _error.AnalysisError("Non-analyzable state initializer");
      }
      const stateName = (0, _utils.memberName)(fieldPath.node);
      if (stateName == null) {
        throw new _error.AnalysisError("Non-analyzable state initializer");
      }
      const fieldInitPath = fieldPath.get("value");
      if (!fieldInitPath.isExpression()) {
        throw new _error.AnalysisError("Non-analyzable state initializer");
      }
      const state = getState(stateName);
      state.sites.push({
        type: "state_init",
        path: fieldPath,
        valuePath: fieldInitPath
      });
    }
  }
  for (const site of stateObjSites) {
    if (site.init) {
      continue;
    }
    if (site.type !== "expr" || site.hasWrite) {
      throw new _error.AnalysisError(`Invalid use of this.state`);
    }
    const gpPath = site.path.parentPath;
    if (!gpPath.isMemberExpression()) {
      throw new _error.AnalysisError(`Stray this.state`);
    }
    const stateName = (0, _utils.memberRefName)(gpPath.node);
    if (stateName == null) {
      throw new _error.AnalysisError(`Non-analyzable state name`);
    }
    const state = getState(stateName);
    state.sites.push({
      type: "expr",
      path: gpPath
    });
  }
  for (const site of setStateSites) {
    if (site.type !== "expr" || site.hasWrite) {
      throw new _error.AnalysisError(`Invalid use of this.setState`);
    }
    const gpPath = site.path.parentPath;
    if (!gpPath.isCallExpression()) {
      throw new _error.AnalysisError(`Stray this.setState`);
    }
    const args = gpPath.get("arguments");
    if (args.length !== 1) {
      throw new _error.AnalysisError(`Non-analyzable setState`);
    }
    const arg0 = args[0];
    if (arg0.isObjectExpression()) {
      const props = arg0.get("properties");
      if (props.length !== 1) {
        throw new _error.AnalysisError(`Multiple assignments in setState is not yet supported`);
      }
      const prop0 = props[0];
      if (!prop0.isObjectProperty()) {
        throw new _error.AnalysisError(`Non-analyzable setState`);
      }
      const setStateName = (0, _utils.memberName)(prop0.node);
      if (setStateName == null) {
        throw new _error.AnalysisError(`Non-analyzable setState name`);
      }
      const state = getState(setStateName);
      state.sites.push({
        type: "setState",
        path: gpPath,
        valuePath: prop0.get("value")
      });
    } else {
      throw new _error.AnalysisError(`Non-analyzable setState`);
    }
  }
  for (const [name, state] of states.entries()) {
    const numInits = state.sites.reduce((n, site) => n + Number(site.type === "state_init"), 0);
    if (numInits > 1) {
      throw new _error.AnalysisError(`${name} is initialized more than once`);
    }
    state.init = state.sites.find(site => site.type === "state_init");
  }
  return states;
}